/* Miscellaneous functions

   OperStats - Copyright (c) 2000-2001 Darcy Grexton
   Contact: cygnus@uniserve.ca, skold @ HabberNet, ICQ #6285178

   This program is free but copyrighted software; see the file COPYING
   for details.
 */

#include "../inc/operstats.h"

/* Open the logfile */
void open_log ()
{
    if (!(freopen (LOGFILE, "a", stderr)))
	fatal ("Cannot open logfile! Is the disk full?");
}

/* Log something */
void log (const char *fmt,...)
{
    va_list args;
    time_t t;
    struct tm tm;
    char buf[32];

    va_start (args, fmt);
    time (&t);
    tm = *localtime (&t);
    strftime (buf, sizeof (buf) - 1, "[%d/%m %H:%M:%S] ", &tm);
    
    fputs (buf, stderr);
    vfprintf (stderr, fmt, args);
    fputc ('\n', stderr);
    fflush (stderr);
}

/* Close the logfile */
void close_log ()
{
    fclose (stderr);
}

/* Fatal error: Log it, globop it if we're still connected, then die. */
void fatal (const char *fmt,...)
{
    va_list args;
    char buf[512];

    va_start (args, fmt);    
    vsnprintf (buf, sizeof (buf), fmt, args);

    log ("Fatal error: %s", buf);

    if (servsock > 0)
	globops ("Fatal Error: %s", buf);

    exit (1);
}

/* Send a command to the server.  The two forms here are like
   printf()/vprintf() and friends.
 */
void send_cmd (const char *source, const char *fmt,...)
{
    va_list args;
   
    va_start (args, fmt);
    vsend_cmd (source, fmt, args);
    va_end (args);
}

void vsend_cmd (const char *source, const char *fmt, va_list args)
{
    char buf[2048];
    
    vsnprintf (buf, sizeof (buf), fmt, args);
    if (source)
    {
	sockprintf (servsock, ":%s %s\r\n", source, buf);
	if (debuglevel == 1 || debuglevel == 3)
	    debug ("<-- %s %s", source, buf);
    }
    else
    {
	sockprintf (servsock, "%s\r\n", buf);
	if (debuglevel == 1 || debuglevel == 3)
	    debug ("<-- %s", buf);
    }
}

/* Send a globops */
void globops (const char *sender, const char *fmt,...)
{
    va_list args;
    char buf[2048];
     
    va_start (args, fmt);
    snprintf (buf, sizeof (buf), "GLOBOPS :%s", fmt);

    vsend_cmd (sender ? sender : me.name, buf, args);
    va_end (args);
}

/* PRIVMSG our snooping channel */
void snoop (const char *source, const char *fmt,...) 
{
    va_list args;
    char buf[2048];
    time_t t;
    struct tm tm;
    char timebuf[256];

    if (findchan (snoopchan))
    {
	time (&t);
	tm = *localtime (&t);
	strftime (timebuf, sizeof (timebuf) - 1, "[%H:%M:%S]", &tm);
	va_start (args, fmt);
	snprintf (buf, sizeof (buf), "PRIVMSG %s :%s %s", snoopchan,
	    timebuf, fmt);
	vsend_cmd (source, buf, args);
	va_end (args);
    }
}

/* Send a notice to a user */
void notice (const char *source, const char *dest, const char *fmt,...)
{
    va_list args;
    char buf[2048];

    va_start (args, fmt);
 
    snprintf (buf, sizeof (buf), "NOTICE %s :%s", dest, fmt);
    vsend_cmd (source, buf, args);
    va_end (args);
}

/* Reply to a CTCP command. */
void ctcpreply (const char *source, const char *dest, const char *fmt,...)
{
    va_list args;
    char buf[2048];

    va_start (args, fmt);
    snprintf (buf, sizeof (buf), "NOTICE %s :\001%s\001", dest, fmt);
    vsend_cmd (source, buf, args);
    va_end (args);
}

/* Send a global notice to all users. */
void noticeall (const char *fmt,...)
{
    char *tmp = strtok (tlds, " ");

    /* Is the GlobalNoticer on? Bail out if not.
       (We could easily use OperServ for this, but if people don't have
       the GlobalNoticer on, they probably use something else for notices.)
     */
    if (globalnoticer_on == FALSE)
	return;

#ifdef UNREAL3
    /* Unreal lets us notice $*, so we'll do that. */
    send_cmd (s_GlobalNoticer, "NOTICE $* :%s", fmt);
#else
    
    /* Notice the first TLD */
    send_cmd (s_GlobalNoticer, "NOTICE $*.%s :%s", tmp, fmt);
        
    /* Now run through the list and notice the remaining tlds.
       This will terminate when we run out of tlds..
     */
    for (;;)
    {
	tmp = strtok (NULL, " ");

	if (tmp)
	    send_cmd (s_GlobalNoticer, "NOTICE $*.%s :%s", tmp, fmt);
	else
	    break;
    }
#endif /* UNREAL3 */
}       

/* Write to our debug log */
void debug (const char *fmt,...)
{
    FILE *f;
    va_list args;
    time_t t;
    struct tm tm;
    char buf[256];
 
    va_start (args, fmt);
    time (&t);
    tm = *localtime (&t);
    strftime (buf, sizeof (buf) - 1, "[%d/%m %H:%M:%S] ", &tm);

    /* Name is static for support purposes.
       Opening/closing it isn't pretty, but we probably wont be in debug
       mode long/often enough to care.
     */
    f = fopen ("debug.log", "a");

    /* Can't open it .. hrm. ah well. */
    if (!f)
	return;

    fputs (buf, f);
    vfprintf (f, fmt, args);
    fputc ('\n', f);
    fflush (f);
    fclose (f);
}

/* Case insensitive version of strcmp */
int stricmp (const char *s1, const char *s2)
{
    register int c;

    while ((c = tolower (*s1)) == tolower (*s2))
    {
	if (c == 0)
	    return 0;
	s1++;
	s2++;
    }
    if (c < tolower (*s2))
	return -1; 
    return 1;
}

/* strscpy:  Copy at most len-1 characters from a string to a buffer, and
             add a null terminator after the last character copied.
 */
char * strscpy (char *d, const char *s, size_t len)
{
    char *d_orig = d;

    if (!len)
	return d;
    while (--len && (*d++ = *s++))
	;
    *d = 0;
    return d_orig;
}

void *smalloc (long size)
{
    void *buf;

    buf = malloc (size);
    if (!buf)
        raise (SIGUSR1);
    return buf;
}  

void *scalloc (long elsize, long els)  
{
    void *buf = calloc (elsize, els);
    if (!buf)
	raise (SIGUSR1);
    return buf;
}

void *srealloc (void *oldptr, long newsize)
{   
    void *buf = realloc (oldptr, newsize);
    if (!buf)
        raise (SIGUSR1);
    return buf;
}

char *sstrdup (const char *s)
{   
    char *t = strdup (s);
    if (!t)
        raise (SIGUSR1);
    return t;
}

Hash *get_hash (const char *whoami, const char *source, const char *cmd,
		Hash * hash_table)
{
    for (; hash_table->accept; ++hash_table)
	if (match_wild_nocase (hash_table->accept, cmd))
	{
	    if ((hash_table->access == H_IRCOP && is_oper (source)) ||
		(hash_table->access == H_SRA && is_sra (source)) ||
		(hash_table->access == H_NONE))
		return hash_table;
	    else if ((hash_table->access == H_IRCOP && !is_oper (source)) ||
		     (hash_table->access == H_SRA && !is_sra (source)))
	    {
		notice (whoami, source, ERR_ACCESS_DENIED, cmd);
		return NULL;
	    }
	}
    notice (whoami, source, ERR_UNKNOWN_COMMAND, cmd, whoami);
    return NULL;
}

Hash_HELP *get_help_hash (const char *whoami, const char *source,
			  const char *cmd, Hash_HELP * hash_table)
{
    for (; hash_table->accept; ++hash_table)
	if (match_wild_nocase (hash_table->accept, cmd))
	{
	    if ((hash_table->access == H_IRCOP && is_oper (source)) ||
		(hash_table->access == H_SRA && is_sra (source)) ||
		(hash_table->access == H_NONE))
		return hash_table;
	    else if ((hash_table->access == H_IRCOP && !is_oper (source)) ||
		     (hash_table->access == H_SRA && !is_sra (source)))
	    {
		notice (whoami, source, ERR_ACCESS_DENIED, cmd);
		return NULL;
	    }
	}
    notice (whoami, source, ERR_NO_HELP, cmd);
    return NULL;
}

/* Is the given nick an oper? */
int is_oper (const char *nick)  
{
    User *user = finduser (nick);

    return (user && (user->mode & UMODE_o));
}

/* Is the given nick an SRA? */
int is_sra (const char *nick)
{
    User *user = finduser (nick);

    return (user && (user->flags & UF_SRA));
}

/* Search case-insensitively for string s2 within string s1 */
char *stristr (char *s1, char *s2)
{ 
    register char *s = s1, *d = s2;

    while (*s1)
	if (tolower (*s1) == tolower (*d))
	{
	    s1++;
	    d++; 
	    if (*d == 0)
		return s;
	}
	else
	{
	    s = ++s1;
	    d = s2;  
	}
    return NULL;
}

static int do_match_wild (const char *pattern, const char *str, int  docase)
{
    char c;
    const char *s;

    for (;;)
    {
	switch (c = *pattern++)
	{
	    case 0:
		if (!*str)
		    return 1;
		return 0;
	    case '?':
		if (!*str)
		    return 0;
		str++;
		break;
	    case '*':
		if (!*pattern)
		    return 1;
		s = str;
		while (*s)
		{
		    if ((docase ? (*s==*pattern) : (tolower (*s)==tolower
			(*pattern))) && do_match_wild (pattern, s, docase))
			return 1;
		    s++;
		}
		break;
	    default:
		if (docase ? (*str++ != c) : (tolower(*str++) != tolower(c)))
		    return 0;
		break;
	}
    }
}

int match_wild (const char *pattern, const char *str)
{
    return do_match_wild (pattern, str, 1);
}

int match_wild_nocase (const char *pattern, const char *str)
{
    return do_match_wild (pattern, str, 0);
}

/* strupper: Convert a string to upper case. */
char *strupper (char *s)
{
    char *t = s;
    while (*t)
	*t++ = toupper (*t);
    return s;
}

/* strlower: Convert a string to lower case. */
char *strlower (char *s)
{
    char *t = s;
    while (*t)
	*t++ = tolower (*t);
    return s;
}

/* Strip unwanted chars out of lines - Borrowed from GeoStats */
void strip (char *line)
{  
    char *c;

    if (line)
    {
	if ((c = strchr (line, '\n')))
	    *c = '\0';
	if ((c = strchr (line, '\r')))
	    *c = '\0';
	if ((c = strchr (line, '\1')))
	    *c = '\0';
    }
}

/* Is it one of ours? */
int is_one_of_mine (const char *nick)
{
    if ((operserv_on == TRUE && !stricmp (nick, s_OperServ)) ||
        (statserv_on == TRUE && !stricmp (nick, s_StatServ)) ||
	(globalnoticer_on == TRUE && !stricmp (nick, s_GlobalNoticer)))
        return 1;
 
    return 0;
}

/* Return the date an event happened */
char *get_time (time_t event)
{
    struct tm tm;
    static char timebuf[256];

    tm = *gmtime (&event);
    strftime (timebuf, sizeof (timebuf), "%a, %b %d, %Y, %H:%M:%S GMT", &tm);
    timebuf[sizeof(timebuf)-1] = 0;
    return timebuf;
}

/* Return the time elapsed since an event */
char *time_ago (time_t event)
{
    static char ret[128];
    int years, weeks, days, hours, minutes, seconds;

    event = time (NULL) - event;
    years = weeks = days = hours = minutes = seconds = 0;

    while (event >= 60 * 60 * 24 * 365)
    {
	event -= 60 * 60 * 24 * 365;
	years++;
    }
    while (event >= 60 * 60 * 24 * 7)
    {
	event -= 60 * 60 * 24 * 7;
	weeks++;
    }
    while (event >= 60 * 60 * 24)
    {
	event -= 60 * 60 * 24;
	days++; 
    }
    while (event >= 60 * 60)
    {
	event -= 60 * 60;
	hours++;
    }
    while (event >= 60)
    {
	event -= 60;
	minutes++;
    }

    seconds = event;
     
    if (years)
	snprintf (ret, sizeof (ret),
		"%d year%s, %d week%s, %d day%s, %02d:%02d:%02d",
		years, years == 1 ? "" : "s",
		weeks, weeks == 1 ? "" : "s",
		days, days == 1 ? "" : "s",
		hours, minutes, seconds);
    else if (weeks)
	snprintf (ret, sizeof (ret), "%d week%s, %d day%s, %02d:%02d:%02d",
		weeks, weeks == 1 ? "" : "s",
		days, days == 1 ? "" : "s",
		hours, minutes, seconds);
    else if (days)
	snprintf (ret, sizeof (ret), "%d day%s, %02d:%02d:%02d",
		days, days == 1 ? "" : "s",  
		hours, minutes, seconds);
    else if (hours)
	snprintf (ret, sizeof (ret), "%d hour%s, %d minute%s, %d second%s",
		hours, hours == 1 ? "" : "s",
		minutes, minutes == 1 ? "" : "s",
		seconds, seconds == 1 ? "" : "s");
    else if (minutes)
	snprintf (ret, sizeof (ret), "%d minute%s, %d second%s",
		minutes, minutes == 1 ? "" : "s",
		seconds, seconds == 1 ? "" : "s");
    else
	snprintf (ret, sizeof (ret), "%d second%s",
		seconds, seconds == 1 ? "" : "s");

    return ret;
}

/* Duration: How long since an event happened. 
   We return a time in the same format as we accept it in the conf,
   ie 5d10m = 5 days, 10 minutes .. Like dotime, we only do days,
   hours, minutes and seconds here.
 */
char *duration (time_t event)
{
    static char ret[128];
    int days, hours, minutes, seconds;

    days = hours = minutes = seconds = 0;

    while (event >= 60 * 60 * 24)
    {
	event -= 60 * 60 * 24;
	days++; 
    }
    while (event >= 60 * 60)
    {
	event -= 60 * 60;
	hours++;
    }
    while (event >= 60)
    {
	event -= 60;
	minutes++;
    }

    seconds = event;

    /* This is pretty big, but it works.. */
     
    if (days && hours && minutes && seconds)
	snprintf (ret, sizeof (ret),
	    "%dd%dh%dm%ds", days, hours, minutes, seconds);
    else if (days && hours && minutes)
	snprintf (ret, sizeof (ret),
	    "%dd%dh%dm", days, hours, minutes);
    else if (days && hours)
	snprintf (ret, sizeof (ret),
	    "%dd%dh", days, hours);
    else if (days)
	snprintf (ret, sizeof (ret),
	    "%dd", days);
    else if (hours && minutes && seconds)
	snprintf (ret, sizeof (ret),
	    "%dh%dm%ds", hours, minutes, seconds);
    else if (hours && minutes)
	snprintf (ret, sizeof (ret),
	    "%dh%dm", hours, minutes);
    else if (hours)
	snprintf (ret, sizeof (ret),
	    "%dh", hours);
    else if (minutes && seconds)
	snprintf (ret, sizeof (ret),
	    "%dm%ds", minutes, seconds);
    else if (minutes)
	snprintf (ret, sizeof (ret), 
	    "%dm", minutes);
    else
	snprintf (ret, sizeof (ret),
	    "%ds", seconds);

    return ret;
}

/* dotime:  Return the number of seconds corresponding to the given time
            string.  If the given string does not represent a valid time,
            return -1.
  
            A time string is either a plain integer (representing a number
            of seconds), or an integer followed by one of these characters:
            "s" (seconds), "m" (minutes), "h" (hours), or "d" (days).

   Borrowed from IRCServices, modified to accept stacking of times. -skold
 */
int dotime (const char *s)
{
    int amount, seconds = 0;

    amount = strtol (s, (char **)&s, 10);

    while (*s)
    {
	switch (*s++)
	{
	    case 's':
		seconds += amount;
		amount = strtol (s, (char **)&s, 10);
		break;
	    case 'm':
		seconds += amount * 60;
		amount = strtol (s, (char **)&s, 10);
		break;
	    case 'h':
		seconds += amount * 3600;
		amount = strtol (s, (char **)&s, 10);
		break;
	    case 'd':
		seconds += amount * 86400;
		amount = strtol (s, (char **)&s, 10);
		break;
	}
    }
    return seconds;
}

/* Send a NULL-terminated array of text as NOTICEs. */
void notice_list (const char *source, const char *dest, const char **text)
{
    while (*text)
    {
	if (**text)
	    notice (source, dest, *text);
	else
	    notice (source, dest, " ");
	text++;  
    }
}

/* Find an SRA by nick. Return NULL if sra could not be found. */
SRA *findsra (const char *nick)
{
    SRA *sra;

    if (!nick)  
	return NULL;

    for (sra = sralist; sra; sra = sra->next)
	if (!stricmp (nick, sra->nick))
	    return sra;

    return NULL;
}

/* Find an TLD. Return NULL if tld could not be found. */
TLD *findtld (const char *param)
{
    TLD *tld;

    if (!param)
	return NULL;

    for (tld = tldlist; tld; tld = tld->next)
	if (!stricmp (param, tld->name))
	    return tld;

    return NULL;
}

/* Find a server for stats purposes. */
ServStat *findserv (const char *server)
{
    ServStat *serverstat;

    serverstat = statlist[SHASH(server)];
    while (serverstat && stricmp (serverstat->name, server))
	serverstat = serverstat->next;
    return serverstat;
}

/* Is the given mask akilled? */
int is_akilled (const char *mask)
{   
    AutoKill *akill;
    int i;
        
    for (i = 0; i < 1024; i++)
	for (akill = akilllist[i]; akill; akill = akill->next)
	    if (match_wild_nocase (akill->mask, mask))
		return 1;
    return 0;
}

/* Given a user, return a mask that will most likely match any address the
   user will have from that location.  For IP addresses, wildcards the
   appropriate subnet mask (e.g. 35.1.1.1 -> 35.*; 128.2.1.1 -> 128.2.*);
   for named addresses, wildcards the leftmost part of the name unless the
   name only contains two parts.  If the username begins with a ~, delete
   it.  The returned character string is malloc'd and should be free'd
   when done with.

   Borrowed from IRCServices.
 */
char *create_mask (User *u)
{
    char *mask, *s, *end;

    end = mask = smalloc (strlen (u->user) + strlen (u->host) + 2);
    end += sprintf (end, "*@");
    if (strspn (u->host, "0123456789.") == strlen (u->host)
	&& (s = strchr(u->host, '.')) && (s = strchr(s+1, '.'))
	&& (s = strchr(s+1, '.'))  && (   !strchr(s+1, '.')))
    {
	s = sstrdup (u->host);
	*strrchr (s, '.') = 0;
	if (atoi (u->host) < 192)
	    *strrchr (s, '.') = 0;
	if (atoi (u->host) < 128)
	    *strrchr (s, '.') = 0;
	sprintf (end, "%s.*", s);
	free(s);
    }
    else
    {
	if ((s = strchr (u->host, '.')) && strchr (s+1, '.'))
	{
	    s = sstrdup (strchr (u->host, '.') -1);
	    *s = '*';
	}
	else
	    s = sstrdup (u->host);

	strcpy (end, s);
	free (s);
    }
    return mask;
}
